@@ -10,19 +10,18 @@ This module provides the core HttpBin experience.
 import base64
 import json
 import os
+import random
 import time
 import uuid
-import random
-import base64
 
-from flask import Flask, Response, request, render_template, redirect, jsonify, make_response
-from werkzeug.datastructures import WWWAuthenticate
+from flask import Flask, Response, request, render_template, redirect, jsonify as flask_jsonify, make_response, url_for
+from werkzeug.datastructures import WWWAuthenticate, MultiDict
 from werkzeug.http import http_date
 from werkzeug.wrappers import BaseResponse
 from six.moves import range as xrange
 
 from . import filters
-from .helpers import get_headers, status_code, get_dict, check_basic_auth, check_digest_auth, H, ROBOT_TXT, ANGRY_ASCII
+from .helpers import get_headers, status_code, get_dict, get_request_range, check_basic_auth, check_digest_auth, secure_cookie, H, ROBOT_TXT, ANGRY_ASCII
 from .utils import weighted_choice
 from .structures import CaseInsensitiveDict
 
@@ -37,6 +36,12 @@ ENV_COOKIES = (
     '__utmb'
 )
 
+def jsonify(*args, **kwargs):
+    response = flask_jsonify(*args, **kwargs)
+    if not response.data.endswith(b'\n'):
+        response.data += b'\n'
+    return response
+
 # Prevent WSGI from correcting the casing of the Location header
 BaseResponse.autocorrect_location_header = False
 
@@ -45,6 +50,23 @@ tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
 
 app = Flask(__name__, template_folder=tmpl_dir)
 
+# Set up Bugsnag exception tracking, if desired. To use Bugsnag, install the
+# Bugsnag Python client with the command "pip install bugsnag", and set the
+# environment variable BUGSNAG_API_KEY. You can also optionally set
+# BUGSNAG_RELEASE_STAGE.
+if os.environ.get("BUGSNAG_API_KEY") is not None:
+    try:
+        import bugsnag
+        import bugsnag.flask
+        release_stage = os.environ.get("BUGSNAG_RELEASE_STAGE") or "production"
+        bugsnag.configure(api_key=os.environ.get("BUGSNAG_API_KEY"),
+                          project_root=os.path.dirname(os.path.abspath(__file__)),
+                          use_ssl=True, release_stage=release_stage,
+                          ignore_classes=['werkzeug.exceptions.NotFound'])
+        bugsnag.flask.handle_exceptions(app)
+    except:
+        app.logger.warning("Unable to initialize Bugsnag exception handling.")
+
 # -----------
 # Middlewares
 # -----------
@@ -58,6 +80,8 @@ def set_cors_headers(response):
         # http://www.w3.org/TR/cors/#access-control-allow-methods-response-header
         response.headers['Access-Control-Allow-Methods'] = 'GET, POST, PUT, DELETE, PATCH, OPTIONS'
         response.headers['Access-Control-Max-Age'] = '3600'  # 1 hour cache
+        if request.headers.get('Access-Control-Request-Headers') is not None:
+            response.headers['Access-Control-Allow-Headers'] = request.headers['Access-Control-Request-Headers']
     return response
 
 
@@ -68,8 +92,8 @@ def set_cors_headers(response):
 @app.route('/')
 def view_landing_page():
     """Generates Landing Page."""
-
-    return render_template('index.html')
+    tracking_enabled = 'HTTPBIN_TRACKING' in os.environ
+    return render_template('index.html', tracking_enabled=tracking_enabled)
 
 
 @app.route('/html')
@@ -155,7 +179,7 @@ def view_patch():
 
 @app.route('/delete', methods=('DELETE',))
 def view_delete():
-    """Returns DETLETE Data."""
+    """Returns DELETE Data."""
 
     return jsonify(get_dict(
         'url', 'args', 'form', 'data', 'origin', 'headers', 'files', 'json'))
@@ -181,14 +205,22 @@ def view_deflate_encoded_content():
 
 @app.route('/redirect/<int:n>')
 def redirect_n_times(n):
-    """301 Redirects n times."""
-
+    """302 Redirects n times."""
     assert n > 0
 
-    if (n == 1):
-        return redirect('/get')
+    absolute = request.args.get('absolute', 'false').lower() == 'true'
 
-    return redirect('/redirect/{0}'.format(n - 1))
+    if n == 1:
+        return redirect(url_for('view_get', _external=absolute))
+
+    if absolute:
+        return _redirect('absolute', n, True)
+    else:
+        return _redirect('relative', n, False)
+
+
+def _redirect(kind, n, external):
+    return redirect(url_for('{0}_redirect_n_times'.format(kind), n=n - 1, _external=external))
 
 
 @app.route('/redirect-to')
@@ -209,21 +241,33 @@ def redirect_to():
 
 @app.route('/relative-redirect/<int:n>')
 def relative_redirect_n_times(n):
-    """301 Redirects n times."""
+    """302 Redirects n times."""
 
     assert n > 0
 
     response = app.make_response('')
     response.status_code = 302
 
-    if (n == 1):
-        response.headers['Location'] = '/get'
+    if n == 1:
+        response.headers['Location'] = url_for('view_get')
         return response
 
-    response.headers['Location'] = '/relative-redirect/{0}'.format(n - 1)
+    response.headers['Location'] = url_for('relative_redirect_n_times', n=n - 1)
     return response
 
 
+@app.route('/absolute-redirect/<int:n>')
+def absolute_redirect_n_times(n):
+    """302 Redirects n times."""
+
+    assert n > 0
+
+    if n == 1:
+        return redirect(url_for('view_get', _external=True))
+
+    return _redirect('absolute', n, True)
+
+
 @app.route('/stream/<int:n>')
 def stream_n_messages(n):
     """Stream n JSON messages"""
@@ -236,7 +280,6 @@ def stream_n_messages(n):
             yield json.dumps(response) + '\n'
 
     return Response(generate_stream(), headers={
-        "Transfer-Encoding": "chunked",
         "Content-Type": "application/json",
         })
 
@@ -267,14 +310,20 @@ def view_status_code(codes):
 @app.route('/response-headers')
 def response_headers():
     """Returns a set of response headers from the query string """
-    headers = CaseInsensitiveDict(request.args.items())
-    response = jsonify(headers.items())
+    headers = MultiDict(request.args.items(multi=True))
+    response = jsonify(headers.lists())
 
     while True:
         content_len_shown = response.headers['Content-Length']
-        response = jsonify(response.headers.items())
-        for key, value in headers.items():
-            response.headers[key] = value
+        d = {}
+        for key in response.headers.keys():
+            value = response.headers.get_all(key)
+            if len(value) == 1:
+                value = value[0]
+            d[key] = value
+        response = jsonify(d)
+        for key, value in headers.items(multi=True):
+            response.headers.add(key, value)
         if response.headers['Content-Length'] == content_len_shown:
             break
     return response
@@ -307,8 +356,8 @@ def view_forms_post():
 def set_cookie(name, value):
     """Sets a cookie and redirects to cookie list."""
 
-    r = app.make_response(redirect('/cookies'))
-    r.set_cookie(key=name, value=value)
+    r = app.make_response(redirect(url_for('view_cookies')))
+    r.set_cookie(key=name, value=value, secure=secure_cookie())
 
     return r
 
@@ -318,9 +367,9 @@ def set_cookies():
     """Sets cookie(s) as provided by the query string and redirects to cookie list."""
 
     cookies = dict(request.args.items())
-    r = app.make_response(redirect('/cookies'))
+    r = app.make_response(redirect(url_for('view_cookies')))
     for key, value in cookies.items():
-        r.set_cookie(key=key, value=value)
+        r.set_cookie(key=key, value=value, secure=secure_cookie())
 
     return r
 
@@ -330,7 +379,7 @@ def delete_cookies():
     """Deletes cookie(s) as provided by the query string and redirects to cookie list."""
 
     cookies = dict(request.args.items())
-    r = app.make_response(redirect('/cookies'))
+    r = app.make_response(redirect(url_for('view_cookies')))
     for key, value in cookies.items():
         r.delete_cookie(key=key)
 
@@ -405,6 +454,7 @@ def drip():
     args = CaseInsensitiveDict(request.args.items())
     duration = float(args.get('duration', 2))
     numbytes = int(args.get('numbytes', 10))
+    code = int(args.get('code', 200))
     pause = duration / numbytes
 
     delay = float(args.get('delay', 0))
@@ -416,9 +466,14 @@ def drip():
             yield u"*".encode('utf-8')
             time.sleep(pause)
 
-    return Response(generate_bytes(), headers={
+    response = Response(generate_bytes(), headers={
         "Content-Type": "application/octet-stream",
-        })
+        "Content-Length": str(numbytes),
+    })
+
+    response.status_code = code
+
+    return response
 
 @app.route('/base64/<value>')
 def decode_base64(value):
@@ -449,6 +504,11 @@ def cache_control(value):
     return response
 
 
+@app.route('/encoding/utf8')
+def encoding():
+    return render_template('UTF-8-demo.txt')
+
+
 @app.route('/bytes/<int:n>')
 def random_bytes(n):
     """Returns n random bytes generated with given seed."""
@@ -492,25 +552,90 @@ def stream_random_bytes(n):
         if chunks:
             yield(bytes(chunks))
 
-    headers = {'Transfer-Encoding': 'chunked',
-               'Content-Type': 'application/octet-stream'}
+    headers = {'Content-Type': 'application/octet-stream'}
 
     return Response(generate_bytes(), headers=headers)
 
+@app.route('/range/<int:numbytes>')
+def range_request(numbytes):
+    """Streams n random bytes generated with given seed, at given chunk size per packet."""
+
+    if numbytes <= 0 or numbytes > (100 * 1024):
+        response = Response(headers={
+            'ETag' : 'range%d' % numbytes,
+            'Accept-Ranges' : 'bytes'
+            })
+        response.status_code = 404
+        response.data = 'number of bytes must be in the range (0, 10240]'
+        return response
+
+    params = CaseInsensitiveDict(request.args.items())
+    if 'chunk_size' in params:
+        chunk_size = max(1, int(params['chunk_size']))
+    else:
+        chunk_size = 10 * 1024
+
+    duration = float(params.get('duration', 0))
+    pause_per_byte = duration / numbytes
+
+    request_headers = get_headers()
+    first_byte_pos, last_byte_pos = get_request_range(request_headers, numbytes)
+
+    if first_byte_pos > last_byte_pos or first_byte_pos not in xrange(0, numbytes) or last_byte_pos not in xrange(0, numbytes):
+        response = Response(headers={
+            'ETag' : 'range%d' % numbytes,
+            'Accept-Ranges' : 'bytes',
+            'Content-Range' : 'bytes */%d' % numbytes
+            })
+        response.status_code = 416
+        return response
+
+    def generate_bytes():
+        chunks = bytearray()
+
+        for i in xrange(first_byte_pos, last_byte_pos + 1):
+
+            # We don't want the resource to change across requests, so we need
+            # to use a predictable data generation function
+            chunks.append(ord('a') + (i % 26))
+            if len(chunks) == chunk_size:
+                yield(bytes(chunks))
+                time.sleep(pause_per_byte * chunk_size)
+                chunks = bytearray()
+
+        if chunks:
+            time.sleep(pause_per_byte * len(chunks))
+            yield(bytes(chunks))
+
+    content_range = 'bytes %d-%d/%d' % (first_byte_pos, last_byte_pos, numbytes)
+    response_headers = {
+        'Content-Type': 'application/octet-stream',
+        'ETag' : 'range%d' % numbytes,
+        'Accept-Ranges' : 'bytes',
+        'Content-Range' : content_range }
+
+    response = Response(generate_bytes(), headers=response_headers)
+
+    if (first_byte_pos == 0) and (last_byte_pos == (numbytes - 1)):
+        response.status_code = 200
+    else:
+        response.status_code = 206
+
+    return response
 
 @app.route('/links/<int:n>/<int:offset>')
 def link_page(n, offset):
     """Generate a page containing n links to other pages which do the same."""
     n = min(max(1, n), 200) # limit to between 1 and 200 links
 
-    link = "<a href='/links/{0}/{1}'>{2}</a> "
+    link = "<a href='{0}'>{1}</a> "
 
     html = ['<html><head><title>Links</title></head><body>']
     for i in xrange(n):
         if i == offset:
             html.append("{0} ".format(i))
         else:
-            html.append(link.format(n, i, i))
+            html.append(link.format(url_for('link_page', n=n, offset=i), i))
     html.append('</body></html>')
 
     return ''.join(html)
@@ -519,7 +644,7 @@ def link_page(n, offset):
 @app.route('/links/<int:n>')
 def links(n):
     """Redirect to first links page."""
-    return redirect("/links/{0}/0".format(n))
+    return redirect(url_for('link_page', n=n, offset=0))
 
 
 @app.route('/image')
@@ -527,12 +652,52 @@ def image():
     """Returns a simple image of the type suggest by the Accept header."""
 
     headers = get_headers()
-    if headers['accept'].lower() == 'image/png' or headers['accept'].lower() == 'image/*':
-        return Response(base64.b64decode('iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAACklEQVR4nGMAAQAABQABDQottAAAAABJRU5ErkJggg=='), headers={'Content-Type': 'image/png'})
-    elif headers['accept'].lower() == 'image/jpeg':
-        return Response(base64.b64decode('/9j/4AAQSkZJRgABAQEASABIAAD/2wBDAAMCAgICAgMCAgIDAwMDBAYEBAQEBAgGBgUGCQgKCgkICQkKDA8MCgsOCwkJDRENDg8QEBEQCgwSExIQEw8QEBD/yQALCAABAAEBAREA/8wABgAQEAX/2gAIAQEAAD8A0s8g/9k='), headers={'Content-Type': 'image/jpeg'})
+    if 'accept' not in headers:
+        return image_png() # Default media type to png
+
+    accept = headers['accept'].lower()
+
+    if 'image/webp' in accept:
+        return image_webp()
+    elif 'image/svg+xml' in accept:
+        return image_svg()
+    elif 'image/jpeg' in accept:
+        return image_jpeg()
+    elif 'image/png' in accept or 'image/*' in accept:
+        return image_png()
     else:
-        return status_code(404)
+        return status_code(406) # Unsupported media type
+
+
+@app.route('/image/png')
+def image_png():
+    data = resource('images/pig_icon.png')
+    return Response(data, headers={'Content-Type': 'image/png'})
+
+
+@app.route('/image/jpeg')
+def image_jpeg():
+    data = resource('images/jackal.jpg')
+    return Response(data, headers={'Content-Type': 'image/jpeg'})
+
+
+@app.route('/image/webp')
+def image_webp():
+    data = resource('images/wolf_1.webp')
+    return Response(data, headers={'Content-Type': 'image/webp'})
+
+
+@app.route('/image/svg')
+def image_svg():
+    data = resource('images/svg_logo.svg')
+    return Response(data, headers={'Content-Type': 'image/svg+xml'})
+
+
+def resource(filename):
+    path = os.path.join(
+        tmpl_dir,
+        filename)
+    return open(path, 'rb').read()
 
 
 @app.route("/xml")
